"""
The MIT License (MIT)

Copyright (c) [2015-2017] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

The MIT License (MIT)

Copyright (c) 2013 Philipp Rasch

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from ctypes import c_char_p, c_bool, c_int, c_double,\
    c_char, c_void_p, sizeof, \
    Array, create_string_buffer, cast, Structure, \
    string_at

import numpy
import six
from numpy import ctypeslib as numpc
# Collection of supporting functions for wrapper functions
__author__ = 'AndrewAnnex'

errorformat = """
================================================================================

Toolkit version: {tkvsn}

{short} --
{explain}
{long}

{traceback}

================================================================================\
"""


class SpiceyError(Exception):
    """
    SpiceyError wraps CSPICE errors.
    :type value: str
    """
    def __init__(self, value, found=None):
        self.value = value
        self.found = found

    def __str__(self):
        return self.value


def toDoubleVector(x):
    return DoubleArray.from_param(param=x)


def toDoubleMatrix(x):
    return DoubleMatrix.from_param(param=x)


def toIntVector(x):
    return IntArray.from_param(param=x)


def toIntMatrix(x):
    return IntMatrix.from_param(param=x)


def toBoolVector(x):
    return BoolArray.from_param(param=x)


def toPythonString(inString):
    if six.PY2:
        if isinstance(inString, c_char_p):
            return toPythonString(inString.value)
        return string_at(inString).rstrip()
    elif six.PY3:
        if isinstance(inString, c_char_p):
            return toPythonString(inString.value)
        return bytes.decode(string_at(inString), errors="ignore").rstrip()


def emptyCharArray(xLen=None, yLen=None):
    if not yLen:
        yLen = 1
    if not xLen:
        xLen = 1
    if isinstance(xLen, c_int):
        xLen = xLen.value
    if isinstance(yLen, c_int):
        yLen = yLen.value
    return ((c_char * xLen) * yLen)()


def emptyDoubleMatrix(x=3, y=3):
    if isinstance(x, c_int):
        x = x.value
    if isinstance(y, c_int):
        y = y.value
    return ((c_double * x) * y)()


def emptyDoubleVector(n):
    if isinstance(n, c_int):
        n = n.value
    assert(isinstance(n, int))
    return (c_double * n)()


def emptyIntMatrix(x=3, y=3):
    if isinstance(x, c_int):
        x = x.value
    if isinstance(y, c_int):
        y = y.value
    return ((c_int * x) * y)()


def emptyIntVector(n):
    if isinstance(n, c_int):
        n = n.value
    assert (isinstance(n, int))
    return (c_int * n)()


def emptyBoolVector(n):
    if isinstance(n, c_int):
        n = n.value
    return (c_bool * n)()


def cVectorToPython(x):
    """
    Convert the c vector data into the correct python data type
    (numpy arrays or strings)
    :param x:
    :return:
    """
    if isinstance(x[0], bool):
        return numpy.fromiter(x, numpy.bool, count=len(x))
    elif isinstance(x[0], int):
        return numpy.fromiter(x, numpy.int_, count=len(x))
    elif isinstance(x[0], float):
        return numpy.fromiter(x, numpy.float64, count=len(x))
    elif isinstance(x[0].value, bytes):
        return [toPythonString(y) for y in x]


def cMatrixToNumpy(x):
    """
    Convert a ctypes 2d array (or matrix) into a numpy array for python use
    :param x: thing to convert
    :return: numpy.ndarray
    """
    return numpc.as_array(x)


def stringToCharP(inobject, inlen=None):

    """
    :param inobject: input string, int for getting null string of length of int
    :param inlen: optional parameter, length of a given string can be specified
    :return:
    """
    if inlen and isinstance(inobject, str):
        return create_string_buffer(inobject.encode(encoding='UTF-8'), inlen)
    if isinstance(inobject, bytes):
        return inobject
    if isinstance(inobject, c_int):
        return stringToCharP(" " * inobject.value)
    if isinstance(inobject, int):
        return stringToCharP(" " * inobject)
    return c_char_p(inobject.encode(encoding='UTF-8'))


def listToCharArray(inList, xLen=None, yLen=None):
    assert (isinstance(inList, list))
    if not yLen:
        yLen = len(inList)
    if not xLen:
        xLen = max(len(s) for s in inList) + 1
    if isinstance(xLen, c_int):
        xLen = xLen.value
    if isinstance(yLen, c_int):
        yLen = yLen.value
    return ((c_char * xLen) * yLen)(*[stringToCharP(l, inlen=xLen) for l in inList])


def listToCharArrayPtr(inList, xLen=None, yLen=None):
    assert (isinstance(inList, list))
    if not yLen:
        yLen = len(inList)
    if not xLen:
        xLen = max(len(s) for s in inList) + 1
    if isinstance(xLen, c_int):
        xLen = xLen.value
    if isinstance(yLen, c_int):
        yLen = yLen.value
    return cast(((c_char * xLen) * yLen)(*[stringToCharP(l, inlen=xLen) for l in inList]), c_char_p)


class DoubleArrayType:
    # Class type that will handle all double vectors,
    # inspiration from python cookbook 3rd edition
    def from_param(self, param):
        typename = type(param).__name__
        if hasattr(self, 'from_' + typename):
            return getattr(self, 'from_' + typename)(param)
        elif isinstance(param, Array):
            return param
        else:
            raise TypeError("Can't convert %s" % typename)

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_double) * len(param))(*param)
        return val

    # Cast from Tuple
    def from_tuple(self, param):
        val = ((c_double) * len(param))(*param)
        return val

    # Cast from a numpy array,
    def from_ndarray(self, param):
        # return param.data_as(POINTER(c_double))
        # the above older method does not work with
        # functions which take vectors of known size
        return numpy.ctypeslib.as_ctypes(param)

    # Cast from array.array objects
    def from_array(self, param):
        if param.typecode != 'd':
            raise TypeError('must be an array of doubles')
        return self.from_list(param)


class DoubleMatrixType:
    # Class type that will handle all double matricies,
    # inspiration from python cookbook 3rd edition
    def from_param(self, param):
        typename = type(param).__name__
        if hasattr(self, 'from_' + typename):
            return getattr(self, 'from_' + typename)(param)
        elif isinstance(param, Array):
            return param
        else:
            raise TypeError("Can't convert %s" % typename)

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_double * len(param[0])) * len(param))(*[DoubleArray.from_param(x) for x in param])
        return val

    # Cast from Tuple
    def from_tuple(self, param):
        val = ((c_double * len(param[0])) * len(param))(*[DoubleArray.from_param(x) for x in param])
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        return numpy.ctypeslib.as_ctypes(param)

    # Cast from a numpy matrix
    def from_matrix(self, param):
        return numpy.ctypeslib.as_ctypes(param)


class IntArrayType:
    # Class type that will handle all int vectors,
    # inspiration from python cookbook 3rd edition
    def from_param(self, param):
        typename = type(param).__name__
        if hasattr(self, 'from_' + typename):
            return getattr(self, 'from_' + typename)(param)
        elif isinstance(param, Array):
            return param
        else:
            raise TypeError("Can't convert %s" % typename)

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_int) * len(param))(*param)
        return val

    # Cast from Tuple
    def from_tuple(self, param):
        val = ((c_int) * len(param))(*param)
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        # return param.data_as(POINTER(c_int))
        # not sure if long is same as int, it should be..
        # return numpy.ctypeslib.as_ctypes(param)
        return self.from_param(param.tolist())

    # Cast from array.array objects
    def from_array(self, param):
        if param.typecode != 'i':
            raise TypeError('must be an array of ints')
        return self.from_list(param)


class IntMatrixType:
    # Class type that will handle all int matricies,
    # inspiration from python cookbook 3rd edition
    def from_param(self, param):
        typename = type(param).__name__
        if hasattr(self, 'from_' + typename):
            return getattr(self, 'from_' + typename)(param)
        elif isinstance(param, Array):
            return param
        else:
            raise TypeError("Can't convert %s" % typename)

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_int * len(param[0])) * len(param))(*[IntArray.from_param(x) for x in param])
        return val

    # Cast from Tuple
    def from_tuple(self, param):
        val = ((c_int * len(param[0])) * len(param))(*[IntArray.from_param(x) for x in param])
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        return numpy.ctypeslib.as_ctypes(param)

    # Cast from a numpy matrix
    def from_matrix(self, param):
        return numpy.ctypeslib.as_ctypes(param)

class BoolArrayType:
    # Class type that will handle all int vectors,
    # inspiration from python cookbook 3rd edition
    def from_param(self, param):
        typename = type(param).__name__
        if hasattr(self, 'from_' + typename):
            return getattr(self, 'from_' + typename)(param)
        elif isinstance(param, Array):
            return param
        else:
            raise TypeError("Can't convert %s" % typename)

    # Cast from lists/tuples
    def from_list(self, param):
        val = ((c_bool) * len(param))(*param)
        return val

    # Cast from Tuple
    def from_tuple(self, param):
        val = ((c_bool) * len(param))(*param)
        return val

    # Cast from a numpy array
    def from_ndarray(self, param):
        # return param.data_as(POINTER(c_int))
        # not sure if long is same as int, it should be..
        # return numpy.ctypeslib.as_ctypes(param)
        return self.from_param(param.tolist())


DoubleArray = DoubleArrayType()

IntArray = IntArrayType()

IntMatrix = IntMatrixType()

BoolArray = BoolArrayType()

DoubleMatrix = DoubleMatrixType()


class Plane(Structure):
    _fields_ = [
        ('_normal', c_double * 3),
        ('_constant', c_double)
    ]

    @property
    def normal(self):
        return cVectorToPython(self._normal)

    @property
    def constant(self):
        return self._constant

    def __str__(self):
        return '<SpicePlane: normal=%s; constant=%s>' % (', '.join([str(x) for x in self._normal]), self._constant)


class Ellipse(Structure):
    _fields_ = [
        ('_center', c_double * 3),
        ('_semi_major', c_double * 3),
        ('_semi_minor', c_double * 3)
    ]

    @property
    def center(self):
        return cVectorToPython(self._center)

    @property
    def semi_major(self):
        return cVectorToPython(self._semi_major)

    @property
    def semi_minor(self):
        return cVectorToPython(self._semi_minor)

    def __str__(self):
        return '<SpiceEllipse: center = %s, semi_major = %s, semi_minor = %s>' % \
               (self.center, self.semi_major, self.semi_minor)


class DataType(object):
    SPICE_CHR = 0
    SPICE_DP = 1
    SPICE_INT = 2
    SPICE_TIME = 3
    SPICE_BOOL = 4
    CHR = 0
    DP = 1
    INT = 2
    TIME = 3
    BOOL = 4

    def __init__(self):
        pass


class SpiceDSKDescr(Structure):
    _fields_ = [
        ('_surfce', c_int),
        ('_center', c_int),
        ('_dclass', c_int),
        ('_dtype', c_int),
        ('_frmcde', c_int),
        ('_corsys', c_int),
        ('_corpar', c_double * 10),
        ('_co1min', c_double),
        ('_co1max', c_double),
        ('_co2min', c_double),
        ('_co2max', c_double),
        ('_co3min', c_double),
        ('_co3max', c_double),
        ('_start', c_double),
        ('_stop', c_double),
    ]
    @property
    def surfce(self):
        return self._surfce

    @property
    def center(self):
        return self._center

    @property
    def dclass(self):
        return self._dclass

    @property
    def dtype(self):
        return self._dtype

    @property
    def frmcde(self):
        return self._frmcde

    @property
    def corsys(self):
        return self._corsys

    @property
    def corpar(self):
        return cVectorToPython(self._corpar)

    @property
    def co1min(self):
        return self._co1min

    @property
    def co1max(self):
        return self._co1max

    @property
    def co2min(self):
        return self._co2min

    @property
    def co2max(self):
        return self._co2max

    @property
    def co3min(self):
        return self._co3min

    @property
    def co3max(self):
        return self._co3max

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop


class SpiceDLADescr(Structure):
    _fields_ = [
        ('_bwdptr', c_int),
        ('_fwdptr', c_int),
        ('_ibase', c_int),
        ('_isize', c_int),
        ('_dbase', c_int),
        ('_dsize', c_int),
        ('_cbase', c_int),
        ('_csize', c_int)
    ]
    @property
    def bwdptr(self):
        return self._bwdptr

    @property
    def fwdptr(self):
        return self._fwdptr

    @property
    def ibase(self):
        return self._ibase

    @property
    def isize(self):
        return self._isize

    @property
    def dbase(self):
        return self._dbase

    @property
    def dsize(self):
        return self._dsize

    @property
    def cbase(self):
        return self._cbase

    @property
    def csize(self):
        return self._csize


class SpiceEKDataType(c_int):
    _SPICE_CHR = c_int(0)
    _SPICE_DP  = c_int(1)
    _SPICE_INT = c_int(2)
    _SPICE_TIME = c_int(3)
    _SPICE_BOOL = c_int(4)


    _fields_ = [
        ('SPICE_CHR', _SPICE_CHR),
        ('SPICE_DP', _SPICE_DP),
        ('SPICE_INT', _SPICE_INT),
        ('SPICE_TIME', _SPICE_TIME),
        ('SPICE_BOOL', _SPICE_BOOL),
    ]

    SPICE_CHR =  _SPICE_CHR.value
    SPICE_DP  =  _SPICE_DP.value
    SPICE_INT =  _SPICE_INT.value
    SPICE_TIME = _SPICE_TIME.value
    SPICE_BOOL = _SPICE_BOOL.value


def emptySpiceEKDataTypeVector(n):
    if isinstance(n, c_int):
        n = n.value
    assert(isinstance(n, int))
    return (SpiceEKDataType * n)()


class SpiceEKExprClass(c_int):
    _SPICE_EK_EXP_COL = c_int(0)
    _SPICE_EK_EXP_FUNC = c_int(1)
    _SPICE_EK_EXP_EXPR = c_int(2)

    _fields_ = [
        ('SPICE_EK_EXP_COL',  _SPICE_EK_EXP_COL),
        ('SPICE_EK_EXP_FUNC', _SPICE_EK_EXP_FUNC),
        ('SPICE_EK_EXP_EXPR', _SPICE_EK_EXP_EXPR)
    ]

    SPICE_EK_EXP_COL = _SPICE_EK_EXP_COL.value
    SPICE_EK_EXP_FUNC = _SPICE_EK_EXP_FUNC.value
    SPICE_EK_EXP_EXPR = _SPICE_EK_EXP_EXPR.value


class SpiceSPK18Subtype(c_int):
    _S18TP0 = c_int(0)
    _S18TP1 = c_int(1)
    S18TP0  = _S18TP0.value
    S18TP1  = _S18TP1.value
    _fields_ = [
        ('S18TP0', _S18TP0),
        ('S18TP1', _S18TP1)
    ]


def emptySpiceEKExprClassVector(n):
    if isinstance(n, c_int):
        n = n.value
    assert(isinstance(n, int))
    return (SpiceEKExprClass * n)()


class SpiceEKAttDsc(Structure):
    _fields_ = [
        ('_cclass', c_int),
        ('_dtype', SpiceEKDataType),
        ('_strlen', c_int),
        ('_size', c_int),
        ('_indexd', c_bool),
        ('_nullok', c_bool)
    ]

    @property
    def cclass(self):
        return self._cclass

    @property
    def dtype(self):
        return self._dtype.value

    @property
    def strlen(self):
        return self._strlen

    @property
    def size(self):
        return self._size

    @property
    def indexd(self):
        return self._indexd

    @property
    def nullok(self):
        return self._nullok

    def __str__(self):
        return '<SpiceEKAttDsc cclass = %s, dtype = %s, strlen = %s, size = %s, indexd = %s, nullok = %s >' % \
               (self.cclass, self.dtype, self.strlen, self.size, self.indexd, self.nullok)


class SpiceEKSegSum(Structure):
    _fields_ = [
        ('_tabnam', c_char * 65),
        ('_nrows', c_int),
        ('_ncols', c_int),
        ('_cnames', (c_char * 100) * 33),
        ('_cdescrs', SpiceEKAttDsc * 100)
    ]

    @property
    def tabnam(self):
        return toPythonString(self._tabnam)

    @property
    def nrows(self):
        return self._nrows

    @property
    def ncols(self):
        return self._ncols

    @property
    def cnames(self):
        return cVectorToPython(self._cnames)[0:self.ncols]

    @property
    def cdescrs(self):
        return self._cdescrs[0:self.ncols]

    def __str__(self):
        return '<SpiceEKSegSum tabnam = %s, nrows = %s, ncols = %s, cnames = %s, cdescrs = %s >' % (self.tabnam, self.nrows, self.ncols, self.cnames, self.cdescrs)


# SpiceCell implementation below is inpart from github.com/DaRasch/spiceminer/
# and modified as needed for this author, maybe we should work together?

### helper classes/functions ###
BITSIZE = {'char': sizeof(c_char), 'int': sizeof(c_int), 'double': sizeof(c_double), 'bool': sizeof(c_int), 'time': sizeof(c_int)}


def _char_getter(data_p, index, length):
    return toPythonString((c_char * length).from_address(data_p + index * length * BITSIZE['char']))


def _double_getter(data_p, index, length):
    return c_double.from_address(data_p + index * BITSIZE['double']).value


def _int_getter(data_p, index, length):
    return c_int.from_address(data_p + index * BITSIZE['int']).value


def SPICEDOUBLE_CELL(size):
    return SpiceCell.double(size)


def SPICEINT_CELL(size):
    return SpiceCell.integer(size)


def SPICECHAR_CELL(size, length):
    return SpiceCell.character(size, length)

def SPICEBOOL_CELL(size):
    return SpiceCell.bool(size)

def SPICETIME_CELL(size):
    return SpiceCell.time(size)


class SpiceCell(Structure):
    #Most written by DaRasch, see included MIT license at file header
    DATATYPES_ENUM = {'char': 0, 'double': 1, 'int': 2, 'time': 3, 'bool': 4}
    DATATYPES_GET = [_char_getter, _double_getter] + [_int_getter] * 3
    baseSize = 6
    minCharLen = 6
    CTRLBLOCK = 6
    _fields_ = [
        ('dtype', c_int),
        ('length', c_int),
        ('size', c_int),
        ('card', c_int),
        ('isSet', c_int),
        ('adjust', c_int),
        ('init', c_int),
        ('base', c_void_p),
        ('data', c_void_p)
    ]

    def __init__(self, dtype=None, length=None, size=None, card=None, isSet=None, base=None, data=None):
        super(SpiceCell, self).__init__()
        self.dtype = dtype
        self.length = length
        self.size = size
        self.card = card
        self.isSet = isSet
        self.adjust = 0  # Always False, because not implemented
        self.init = 0  # Always False, because this is the constructor
        self.base = base  # void pointer
        self.data = data

    def __str__(self):
        return '<SpiceCell dtype = %s, length = %s, size = %s, card = %s,' \
               ' isSet = %s, adjust = %s, init = %s, base = %s, data = %s>' % \
               (self.dtype, self.length, self.size, self.card, self.isSet,
                self.adjust, self.init, self.base, self.data)

    def is_int(self):
        return self.dtype == 2

    def is_double(self):
        return self.dtype == 1

    def is_char(self):
        return self.dtype == 0

    def is_time(self):
        return self.dtype == 3

    def is_bool(self):
        return self.dtype == 4

    def is_set(self):
        return self.isSet == 1

    @classmethod
    def character(cls, size, length):
        base = (c_char * ((cls.CTRLBLOCK + size) * length))()
        data = (c_char * (size * length)).from_buffer(
            base, cls.CTRLBLOCK * BITSIZE['char'] * length)
        instance = cls(cls.DATATYPES_ENUM['char'], length, size, 0, 1,
                       cast(base, c_void_p),
                       cast(data, c_void_p))
        return instance

    @classmethod
    def integer(cls, size):
        base = (c_int * (cls.CTRLBLOCK + size))()
        data = (c_int * size).from_buffer(
            base, cls.CTRLBLOCK * BITSIZE['int'])
        instance = cls(cls.DATATYPES_ENUM['int'], 0, size, 0, 1,
                       cast(base, c_void_p),
                       cast(data, c_void_p))
        return instance

    @classmethod
    def double(cls, size):
        base = (c_double * (cls.CTRLBLOCK + size))()
        data = (c_double * size).from_buffer(
            base, cls.CTRLBLOCK * BITSIZE['double'])
        instance = cls(cls.DATATYPES_ENUM['double'], 0, size, 0, 1,
                       cast(base, c_void_p),
                       cast(data, c_void_p))
        return instance

    @classmethod
    def bool(cls, size):
        base = (c_int * (cls.CTRLBLOCK + size))()
        data = (c_int * size).from_buffer(
            base, cls.CTRLBLOCK * BITSIZE['bool'])
        instance = cls(cls.DATATYPES_ENUM['bool'], 0, size, 0, 1,
                       cast(base, c_void_p),
                       cast(data, c_void_p))
        return instance

    @classmethod
    def time(cls, size):
        base = (c_int * (cls.CTRLBLOCK + size))()
        data = (c_int * size).from_buffer(
            base, cls.CTRLBLOCK * BITSIZE['time'])
        instance = cls(cls.DATATYPES_ENUM['time'], 0, size, 0, 1,
                       cast(base, c_void_p),
                       cast(data, c_void_p))
        return instance

    def __len__(self):
        return self.card

    def __iter__(self):
        getter = SpiceCell.DATATYPES_GET[self.dtype]
        length, card, data = self.length, self.card, self.data
        for i in six.moves.range(card):
            yield (getter(data, i, length))

    def __contains__(self, key):
        return key in self.__iter__()

    def __getitem__(self, key):
        getter = SpiceCell.DATATYPES_GET[self.dtype]
        if isinstance(key, slice):
            #TODO Typechecking
            if self.card == 0:
                return []
            else:
                start, stop, step = key.indices(self.card)
                return [getter(self.data, i, self.length) for i in six.moves.range(start, stop, step)]
        elif key in six.moves.range(-self.card, self.card):
            index = key if key >= 0 else self.card - abs(key)
            return getter(self.data, index, self.length)
        elif not isinstance(key, int):
            raise TypeError('SpiceCell indices must be integers, not {}'.format(type(key)))
        else:
            raise IndexError('SpiceCell index out of range')

    def reset(self):
        self.card = 0
        self.init = 0



"""
The MIT License (MIT)

Copyright (c) [2015-2017] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import functools
from ctypes import c_bool, c_double, POINTER, CFUNCTYPE, byref

UDFUNS = CFUNCTYPE(None, c_double, POINTER(c_double))
UDFUNB = CFUNCTYPE(None, UDFUNS, c_double, POINTER(c_bool))

def SpiceUDFUNS(f):
    """
    Decorator for wrapping python functions in spice udfuns callback type
    :param f: function that has one argument of type float, and returns a float
    :type f: builtins.function
    :return: wrapped udfunc function
    :rtype: builtins.function
    """

    @functools.wraps(f)
    def wrapping_udfuns(x, value):
        result = f(x)
        value[0] = c_double(result)

    return UDFUNS(wrapping_udfuns)

def SpiceUDFUNB(f):
    """
    Decorator for wrapping python functions in spice udfunb callback type
    :param f:
    :type f: builtins.function
    :return:
    """

    @functools.wraps(f)
    def wrapping_udfunb(udf, et, xbool):
        result = f(udf, et) # the function takes a udffunc as a argument
        xbool[0] = c_bool(result)

    return UDFUNB(wrapping_udfunb)

def CallUDFUNS(f, x):
    """
    We are given a UDF CFUNCTYPE and want to call it in python

    :param f: SpiceUDFUNS
    :type f: CFUNCTYPE
    :param x: some scalar
    :type x: float
    :return: value
    :rtype: float
    """
    value = c_double()
    f(x, byref(value))
    return value.value


"""
The MIT License (MIT)

Copyright (c) [2015-2017] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.test import test as TestCommand
from setuptools.dist import Distribution
import ssl
import sys
import os
import subprocess
import platform
import shutil

__author__ = 'AndrewAnnex'

# Get OS platform
host_OS = platform.system()
# Get platform is Unix-like OS or not
is_unix = host_OS in ("Linux", "Darwin", "FreeBSD")
# Get current working directory
root_dir = os.path.dirname(os.path.realpath(__file__))
# Make the directory path for cspice
cspice_dir = os.path.join(root_dir, 'cspice')
# Make the directory path for cspice/lib
lib_dir = os.path.join(cspice_dir, 'lib')

TEST_DEPENDENCIES = ['numpy>=1.12.0', 'six>=1.9.0', 'pytest>=2.9.0']
DEPENDENCIES = ['numpy>=1.12.0', 'six>=1.9.0', 'certifi>=2017.1.23']
REQUIRES = ['numpy', 'six']

# If we have an old version of OpenSSL, CSPICE will be downloaded
# (if required) using urllib3.  Extend the list of required packages.
if ssl.OPENSSL_VERSION < 'OpenSSL 1.0.1g':
    DEPENDENCIES.extend(['urllib3[secure]>=1.22', 'pyOpenSSL>=17.3.0'])


# py.test integration from pytest.org
class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


class BinaryDistribution(Distribution):
    def is_pure(self):
        return False


class InstallSpiceyPy(install):
    """Class that extends the install command and encapsulates the
    process for installing the required CSPICE distribution at the
    right place.
    """

    def run(self):
        self.check_for_spice()

        print("Host OS: {0}".format(host_OS))
        if is_unix:
            self.unix_method()
        elif host_OS == "Windows":
            self.windows_method()
        else:
            sys.exit("Unsupported OS: {0}".format(host_OS))

        install.run(self)

    @staticmethod
    def check_for_spice():
        print('Checking the path', cspice_dir)
        if not os.path.exists(cspice_dir):
            import getspice
            message = 'Unable to find CSPICE at {0}. Attempting to Download CSPICE For you:'.format(cspice_dir)
            print(message)
            # Download cspice using getspice.py
            getspice.GetCSPICE(version='N0066')
            if not os.path.exists(cspice_dir):
                message = 'Unable to find CSPICE at {0}. Exiting'.format(cspice_dir)
                sys.exit(message)

    @staticmethod
    def unpack_cspice():
        if is_unix:
            cspice_lib = os.path.join(lib_dir, ("cspice.lib" if host_OS is "Windows" else "cspice.a"))
            csupport_lib = os.path.join(lib_dir, ("csupport.lib" if host_OS is "Windows" else "csupport.a"))

            if os.path.exists(cspice_lib) and os.path.exists(csupport_lib):
                cwd = os.getcwd()
                try:
                    os.chdir(lib_dir)
                    if host_OS is "Windows":
                        raise BaseException("Windows is not supported in this build method")
                    elif is_unix:
                        for lib in ["ar -x cspice.a", "ar -x csupport.a"]:
                            unpack_lib_process = subprocess.Popen(lib, shell=True)
                            process_status = os.waitpid(unpack_lib_process.pid, 0)[1]
                            if process_status != 0:
                                raise BaseException('{0}'.format(process_status))
                    else:
                        raise BaseException("Unsupported OS: {0}".format(host_OS))
                except BaseException as error:
                    status = error.args
                    sys.exit('Error: cspice object file extraction failed with exit status: {0}'.format(status))
                finally:
                    os.chdir(cwd)
            else:
                error_Message = "Error, cannot find CSPICE " \
                                "static libraries at {0}".format(lib_dir)
                sys.exit(error_Message)

    @staticmethod
    def build_library():
        # Get the current working directory
        cwd = os.getcwd()

        if is_unix:
            try:
                os.chdir(lib_dir)
                # find a way to make this work via Extension and setuptools, not using popen.
                build_lib = subprocess.Popen('gcc -shared -fPIC -lm *.o -o spice.so', shell=True)
                status = os.waitpid(build_lib.pid, 0)[1]
                if status != 0:
                    raise BaseException('{0}'.format(status))
                success = os.path.exists(os.path.join(os.getcwd(), 'spice.so'))
                if not success:
                    raise BaseException("Did not find spice.so, build went badly.")
            except BaseException as errorInst:
                status = errorInst.args
                sys.exit('Error: compilation of shared spice.so build exit status: {0}'.format(status))

        elif host_OS == "Windows":
            try:
                destination = os.path.join(cspice_dir, "src", "cspice")
                defFile = os.path.join(root_dir, "appveyor", "cspice.def")
                makeBat = os.path.join(root_dir, "appveyor", "makeDynamicSpice.bat")
                shutil.copy(defFile, destination)
                shutil.copy(makeBat, destination)
                # run the script
                os.chdir(destination)
                windows_build = subprocess.Popen("makeDynamicSpice.bat", shell=True)
                status = windows_build.wait()
                if status != 0:
                    raise BaseException('{0}'.format(status))
            except BaseException as error:
                sys.exit("Build failed with: {0}".format(error.args))
        # Change back to the stored 'current working directory
        os.chdir(cwd)

    @staticmethod
    def move_to_root_directory():
        sharedlib = 'spice.so' if is_unix else 'cspice.dll'
        destination = os.path.join(root_dir, 'spiceypy', 'utils', sharedlib)
        if not os.path.isfile(destination):
            if is_unix:
                target = os.path.join(cspice_dir, 'lib', sharedlib)
            else:
                target = os.path.join(cspice_dir, 'src', 'cspice', sharedlib)
            print("Attempting to move: {0}   to: {1}".format(target, destination))
            try:
                os.rename(target, destination)
            except BaseException as e:
                sys.exit('{0} file not found, what happend?: {1}'.format(sharedlib, e))

    @staticmethod
    def cleanup():
        # Remove CSPICE folder
        try:
            shutil.rmtree(os.path.join(os.getcwd(), "cspice"))
        except OSError as e:
            print("Error Cleaning up cspice folder")
            raise e

    def unix_method(self):
        # Unpack cspice.a and csupport.a
        self.unpack_cspice()
        # Build the shared Library
        self.build_library()
        # Move to correct location (root of the distribution)
        self.move_to_root_directory()

    def windows_method(self):
        if os.path.exists(os.path.join(cspice_dir, "lib", "cspice.dll")):
            print("Found pre-made cspice.dll, not building")
        elif os.path.exists(os.path.join(root_dir, 'spiceypy', 'utils', 'cspice.dll')):
            print("Found pre-made cspice.dll in spiceypy, not building")
        else:
            # Build the DLL
            self.build_library()
            # Move to correct location (root of the distribution)
            self.move_to_root_directory()


readme = open('README.rst', 'r')
readmetext = readme.read()
readme.close()

setup(
    name='spiceypy',
    version='2.1.0',
    license='MIT',
    author='Andrew Annex',
    author_email='ama6fy@virginia.edu',
    description='A Python Wrapper for the NAIF CSPICE Toolkit',
    long_description=readmetext,
    keywords=['spiceypy', 'spice', 'naif', 'jpl', 'space', 'geometry'],
    url='https://github.com/AndrewAnnex/SpiceyPy',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: POSIX :: BSD :: FreeBSD",
        "Operating System :: Microsoft :: Windows"
    ],
    packages=find_packages(exclude=["*.tests"]),
    include_package_data=True,
    zip_safe=False,
    distclass=BinaryDistribution,
    package_data={'': ['*.so', "*.dll"]},
    setup_requires=DEPENDENCIES,
    install_requires=DEPENDENCIES,
    requires=REQUIRES,
    tests_require=TEST_DEPENDENCIES,
    cmdclass={
        'install': InstallSpiceyPy,
        'test': PyTest},
    test_suite='spiceypy.tests.test_wrapper.py',
    extras_require={'testing': ['pytest']}
)



"""
The MIT License (MIT)

Copyright (c) [2015-2017] [Andrew Annex]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from ctypes import CDLL, POINTER, c_bool, c_int, c_double, c_char, c_char_p, c_void_p
import os
import platform
from . import support_types as stypes
from . import callbacks

host_OS = platform.system()
sharedLib = "cspice.dll" if host_OS == "Windows" else "spice.so"
sitePath = os.path.join(os.path.dirname(__file__), sharedLib)
libspice = CDLL(sitePath)

__author__ = 'AndrewAnnex'

# ######################################################################################################################
# A

libspice.appndc_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.appndd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.appndi_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.axisar_c.argtypes = [(c_double * 3), c_double, (c_double * 3) * 3]

# #######################################################################################################################
# B
libspice.b1900_c.restype = c_double
libspice.b1950_c.restype = c_double
libspice.bodc2n_c.argtypes = [c_int, c_int, c_char_p, POINTER(c_bool)]
libspice.bodc2s_c.argtypes = [c_int, c_int, c_char_p]
libspice.boddef_c.argtypes = [c_char_p, c_int]
libspice.badkpv_c.argtypes = [c_char_p, c_char_p, c_char_p, c_int, c_int, c_char]
libspice.badkpv_c.restype = c_bool
libspice.bltfrm_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.bodfnd_c.argtypes = [c_int, c_char_p]
libspice.bodfnd_c.restype = c_bool
libspice.bodn2c_c.argtypes = [c_char_p, POINTER(c_int), POINTER(c_bool)]
libspice.bods2c_c.argtypes = [c_char_p, POINTER(c_int), POINTER(c_bool)]
libspice.bodvar_c.argtypes = [c_int, c_char_p, POINTER(c_int), c_void_p]
libspice.bodvcd_c.argtypes = [c_int, c_char_p, c_int, POINTER(c_int), c_void_p]
libspice.bodvrd_c.argtypes = [c_char_p, c_char_p, c_int, POINTER(c_int), c_void_p]
libspice.brcktd_c.argtypes = [c_double, c_double, c_double]
libspice.brcktd_c.restype = c_double
libspice.brckti_c.argtypes = [c_int, c_int, c_int]
libspice.brckti_c.restype = c_int
libspice.bschoc_c.argtypes = [c_char_p, c_int, c_int, c_char_p, POINTER(c_int)]
libspice.bschoc_c.restype = c_int
libspice.bschoi_c.argtypes = [c_int, c_int, POINTER(c_int), POINTER(c_int)]
libspice.bschoi_c.restype = c_int
libspice.bsrchc_c.argtypes = [c_char_p, c_int, c_int, c_char_p]
libspice.bsrchc_c.restype = c_int
libspice.bsrchd_c.argtypes = [c_double, c_int, POINTER(c_double)]
libspice.bsrchd_c.restype = c_int
libspice.bsrchi_c.argtypes = [c_int, c_int, POINTER(c_int)]
libspice.bsrchi_c.restype = c_int
########################################################################################################################
# C
libspice.card_c.argtypes = [POINTER(stypes.SpiceCell)]
libspice.card_c.restype = c_int
libspice.ccifrm_c.argtypes = [c_int, c_int, c_int, POINTER(c_int), c_char_p, POINTER(c_int), POINTER(c_bool)]
libspice.cgv2el_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3), POINTER(stypes.Ellipse)]
libspice.chbder_c.argtypes = [POINTER(c_double), c_int, (c_double*2), c_double, c_int, POINTER(c_double), POINTER(c_double)]
libspice.chkin_c.argtypes = [c_char_p]
libspice.chkout_c.argtypes = [c_char_p]
libspice.cidfrm_c.argtypes = [c_int, c_int, POINTER(c_int), c_char_p, POINTER(c_bool)]
libspice.ckcls_c.argtypes = [c_int]
libspice.ckcov_c.argtypes = [c_char_p, c_int, c_bool, c_char_p, c_double, c_char_p, POINTER(stypes.SpiceCell)]
libspice.ckobj_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.ckgp_c.argtypes = [c_int, c_double, c_double, c_char_p, ((c_double * 3) * 3), POINTER(c_double),
                            POINTER(c_bool)]
libspice.ckgpav_c.argtypes = [c_int, c_double, c_double, c_char_p, ((c_double * 3) * 3), (c_double * 3),
                              POINTER(c_double), POINTER(c_bool)]
libspice.cklpf_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.ckopn_c.argtypes = [c_char_p, c_char_p, c_int, POINTER(c_int)]
libspice.ckupf_c.argtypes = [c_int]
libspice.ckw01_c.argtypes = [c_int, c_double, c_double, c_int, c_char_p, c_bool, c_char_p, c_int, POINTER(c_double),
                             POINTER(c_double * 4), POINTER(c_double * 3)]
libspice.ckw02_c.argtypes = [c_int, c_double, c_double, c_int, c_char_p, c_char_p, c_int, POINTER(c_double),
                             POINTER(c_double), POINTER(c_double * 4), POINTER(c_double * 3), POINTER(c_double)]
libspice.ckw03_c.argtypes = [c_int, c_double, c_double, c_int, c_char_p, c_bool, c_char_p, c_int, POINTER(c_double),
                             POINTER(c_double * 4), POINTER(c_double * 3), c_int, POINTER(c_double)]
libspice.ckw05_c.argtypes = [c_int, c_int, c_int, c_double, c_double, c_int, c_char_p, c_bool, c_char_p, c_int,
                             POINTER(c_double), c_void_p, c_double, c_int, POINTER(c_double)]
libspice.clight_c.argtypes = None
libspice.clight_c.restype = c_double
libspice.clpool_c.argtypes = None
libspice.cltext_.argtypes = [c_char_p, c_int]
libspice.cmprss_c.argtypes = [c_char, c_int, c_char_p, c_int, c_char_p]
libspice.cnmfrm_c.argtypes = [c_char_p, c_int, POINTER(c_int), c_char_p, POINTER(c_bool)]
libspice.conics_c.argtypes = [(c_double * 8), c_double, (c_double * 6)]
libspice.convrt_c.argtypes = [c_double, c_char_p, c_char_p, POINTER(c_double)]
libspice.copy_c.argtypes = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.cpos_c.argtypes = [c_char_p, c_char_p, c_int]
libspice.cpos_c.restype = c_int
libspice.cposr_c.argtypes = [c_char_p, c_char_p, c_int]
libspice.cposr_c.restype = c_int
libspice.cvpool_c.argtypes = [c_char_p, POINTER(c_bool)]
libspice.cyllat_c.argtypes = [c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.cylrec_c.argtypes = [c_double, c_double, c_double, (c_double * 3)]
libspice.cylsph_c.argtypes = [c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]

########################################################################################################################
#D

libspice.dafac_c.argtypes = [c_int, c_int, c_int, c_void_p]
libspice.dafbbs_c.argtypes = [c_int]
libspice.dafbfs_c.argtypes = [c_int]
libspice.dafcls_c.argtypes = [c_int]
libspice.dafcs_c.argtypes = [c_int]
libspice.dafdc_c.argtypes = [c_int]
libspice.dafec_c.argtypes = [c_int, c_int, c_int, POINTER(c_int), c_void_p, POINTER(c_bool)]
libspice.daffna_c.argtypes = [POINTER(c_bool)]
libspice.daffpa_c.argtypes = [POINTER(c_bool)]
libspice.dafgda_c.argtypes = [c_int, c_int, c_int, POINTER(c_double)]
libspice.dafgh_c.argtypes = [POINTER(c_int)]
libspice.dafgn_c.argtypes = [c_int, c_char_p]
libspice.dafgs_c.argtypes = [POINTER(c_double)]
libspice.dafgsr_c.argtypes = [c_int, c_int, c_int, c_int, POINTER(c_double), POINTER(c_bool)]
libspice.dafopr_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.dafopw_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.dafps_c.argtypes = [c_int, c_int, POINTER(c_double), POINTER(c_int), POINTER(c_double)]
libspice.dafrda_c.argtypes = [c_int, c_int, c_int, POINTER(c_double)]
libspice.dafrfr_c.argtypes = [c_int, c_int, POINTER(c_int), POINTER(c_int), c_char_p, POINTER(c_int), POINTER(c_int),
                              POINTER(c_int)]
libspice.dafrs_c.argtype = [POINTER(c_double)]
libspice.dafus_c.argtypes = [POINTER(c_double), c_int, c_int, POINTER(c_double), POINTER(c_int)]
libspice.dasac_c.argtypes = [c_int, c_int, c_int, c_void_p]
libspice.dascls_c.argtypes = [c_int]
libspice.dasdc_c.argtypes = [c_int]
libspice.dasec_c.argtypes = [c_int, c_int, c_int, POINTER(c_int), c_void_p, POINTER(c_bool)]
libspice.dashfn_c.argtypes = [c_int, c_int, c_char_p]
libspice.dasopr_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.dasopw_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.dasonw_.argtypes = [c_char_p, c_char_p, c_char_p, POINTER(c_int), POINTER(c_int), c_int, c_int, c_int]
libspice.dasrfr_c.argtypes = [c_int, c_int, c_int, c_char_p, c_char_p, POINTER(c_int), POINTER(c_int), POINTER(c_int), POINTER(c_int)]
libspice.dcyldr_c.argtypes = [c_double, c_double, c_double, (c_double * 3) * 3]
libspice.deltet_c.argtypes = [c_double, c_char_p, POINTER(c_double)]
libspice.det_c.argtypes = [(c_double * 3) * 3]
libspice.det_c.restype = c_double
libspice.dgeodr_c.argtypes = [c_double, c_double, c_double, c_double, c_double, (c_double * 3) * 3]
libspice.diags2_c.argtypes = [(c_double * 2) * 2, (c_double * 2) * 2, (c_double * 2) * 2]
libspice.diff_c.argtypes = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.dlabbs_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(c_bool)]
libspice.dlabfs_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(c_bool)]
libspice.dlafns_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(stypes.SpiceDLADescr), POINTER(c_bool)]
libspice.dlafps_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(stypes.SpiceDLADescr), POINTER(c_bool)]
libspice.dlatdr_c.argtypes = [c_double, c_double, c_double, (c_double * 3) * 3]
libspice.dp2hx_c.argtypes = [c_double, c_int, c_char_p, POINTER(c_int)]
libspice.dpgrdr_c.argtypes = [c_char_p, c_double, c_double, c_double, c_double, c_double, (c_double * 3) * 3]
libspice.dpmax_c.argtypes = None
libspice.dpmax_c.restype = c_double
libspice.dpmin_c.argtypes = None
libspice.dpmin_c.restype = c_double
libspice.dpr_c.restype = c_double
libspice.drdcyl_c.argtypes = [c_double, c_double, c_double, (c_double * 3) * 3]
libspice.drdgeo_c.argtypes = [c_double, c_double, c_double, c_double, c_double, (c_double * 3) * 3]
libspice.drdlat_c.argtypes = [c_double, c_double, c_double, (c_double * 3) * 3]
libspice.drdpgr_c.argtypes = [c_char_p, c_double, c_double, c_double, c_double, c_double, (c_double * 3) * 3]
libspice.drdsph_c.argtypes = [c_double, c_double, c_double, (c_double * 3) * 3]
libspice.dskb02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(c_int), POINTER(c_int), POINTER(c_int), ((c_double * 3) * 2), POINTER(c_double), (c_double*3), (c_int*3), POINTER(c_int), POINTER(c_int), POINTER(c_int), POINTER(c_int)]
libspice.dskcls_c.argtypes = [c_int, c_bool]
libspice.dskd02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), c_int, c_int, c_int, POINTER(c_int), POINTER(c_double)]
libspice.dskgd_c.argtypes  = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(stypes.SpiceDSKDescr)]
libspice.dskgtl_c.argtypes = [c_int, POINTER(c_double)]
libspice.dski02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), c_int, c_int, c_int, POINTER(c_int), POINTER(c_int)]
libspice.dskmi2_c.argtypes = [c_int, POINTER(c_double * 3), c_int, POINTER(c_int * 3), c_double, c_int, c_int, c_int, c_int, c_bool, c_int, POINTER(c_int * 2), POINTER(c_double), POINTER(c_int)]
libspice.dskn02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), c_int, POINTER(c_double)]
libspice.dskobj_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.dskopn_c.argtypes = [c_char_p, c_char_p, c_int, POINTER(c_int)]
libspice.dskp02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), c_int, c_int, POINTER(c_int), POINTER(c_int * 3)]
libspice.dskrb2_c.argtypes = [c_int, POINTER(c_double * 3), c_int, POINTER(c_int * 3), c_int, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.dsksrf_c.argtypes = [c_char_p, c_int, POINTER(stypes.SpiceCell)]
libspice.dskstl_c.argtypes = [c_int, c_double]
libspice.dskv02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), c_int, c_int, POINTER(c_int), POINTER(c_double * 3)]
libspice.dskw02_c.argtypes = [c_int, c_int, c_int, c_int, c_char_p, c_int, POINTER(c_double), c_double, c_double, c_double, c_double, c_double, c_double, c_double, c_double, c_int, POINTER(c_double * 3), c_int, POINTER(c_int * 3), POINTER(c_double), POINTER(c_int)]
libspice.dskx02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), c_double * 3, c_double * 3, POINTER(c_int), POINTER(c_double), POINTER(c_bool)]
libspice.dskxsi_c.argtypes = [c_bool,  c_char_p, c_int, POINTER(c_int), c_double, c_char_p, c_double * 3, c_double * 3, c_int, c_int, c_double * 3, POINTER(c_int), POINTER(stypes.SpiceDLADescr), POINTER(stypes.SpiceDSKDescr), POINTER(c_double), POINTER(c_int), POINTER(c_bool)]
libspice.dskxv_c.argtypes  = [c_bool, c_char_p, c_int, POINTER(c_int), c_double, c_char_p, c_int, POINTER(c_double *3), POINTER(c_double * 3), POINTER(c_double * 3), POINTER(c_bool)]
libspice.dskz02_c.argtypes = [c_int, POINTER(stypes.SpiceDLADescr), POINTER(c_int), POINTER(c_int)]
libspice.dsphdr_c.argtypes = [c_double, c_double, c_double, (c_double * 3) * 3]
libspice.dtpool_c.argtypes = [c_char_p, POINTER(c_bool), POINTER(c_int), POINTER(c_char)]
libspice.ducrss_c.argtypes = [c_double * 6, c_double * 6, c_double * 6]
libspice.dvcrss_c.argtypes = [c_double * 6, c_double * 6, c_double * 6]
libspice.dvdot_c.argtypes  = [c_double * 6, c_double * 6]
libspice.dvdot_c.restype   = c_double
libspice.dvhat_c.argtypes  = [c_double * 6, c_double * 6]
libspice.dvnorm_c.argtypes = [c_double * 6]
libspice.dvnorm_c.restype  = c_double
libspice.dvpool_c.argtypes = [c_char_p]
libspice.dvsep_c.argtypes  = [c_double * 6, c_double * 6]
libspice.dvsep_c.restype   = c_double
########################################################################################################################
# E

libspice.edlimb_c.argtypes = [c_double, c_double, c_double, (c_double * 3), POINTER(stypes.Ellipse)]
libspice.edterm_c.argtypes = [c_char_p, c_char_p, c_char_p, c_double, c_char_p,
                              c_char_p, c_char_p, c_int, POINTER(c_double),
                              (c_double * 3), c_void_p]
libspice.ekacec_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, c_int, c_void_p, c_int]
libspice.ekaced_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, POINTER(c_double), c_int]
libspice.ekacei_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, POINTER(c_int), c_int]
libspice.ekaclc_c.argtypes = [c_int, c_int, c_char_p, c_int, c_void_p, POINTER(c_int), POINTER(c_bool), POINTER(c_int),
                              POINTER(c_int)]
libspice.ekacld_c.argtypes = [c_int, c_int, c_char_p, POINTER(c_double), POINTER(c_int), POINTER(c_int),
                              POINTER(c_int), POINTER(c_int)]
libspice.ekacli_c.argtypes = [c_int, c_int, c_char_p, POINTER(c_int), POINTER(c_int), POINTER(c_int), POINTER(c_int),
                              POINTER(c_int)]
libspice.ekappr_c.argtypes = [c_int, c_int, POINTER(c_int)]
libspice.ekbseg_c.argtypes = [c_int, c_char_p, c_int, c_int, c_void_p, c_int, c_void_p, POINTER(c_int)]
libspice.ekccnt_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.ekcii_c.argtypes = [c_char_p, c_int, c_int, c_char_p, POINTER(stypes.SpiceEKAttDsc)]
libspice.ekcls_c.argtypes = [c_int]
libspice.ekdelr_c.argtypes = [c_int, c_int, c_int]
libspice.ekffld_c.argtypes = [c_int, c_int, POINTER(c_int)]
libspice.ekfind_c.argtypes = [c_char_p, c_int, POINTER(c_int), POINTER(c_bool), c_char_p]
libspice.ekgc_c.argtypes = [c_int, c_int, c_int, c_int, c_char_p, POINTER(c_bool), POINTER(c_bool)]
libspice.ekgd_c.argtypes = [c_int, c_int, c_int, POINTER(c_double), POINTER(c_bool), POINTER(c_bool)]
libspice.ekgi_c.argtypes = [c_int, c_int, c_int, POINTER(c_int), POINTER(c_bool), POINTER(c_bool)]
libspice.ekifld_c.argtypes = [c_int, c_char_p, c_int, c_int, c_int, c_void_p, c_int, c_void_p, POINTER(c_int),
                              POINTER(c_int)]
libspice.ekinsr_c.argtypes = [c_int, c_int, c_int]
libspice.eklef_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.eknelt_c.argtypes = [c_int, c_int]
libspice.eknelt_c.restype = c_int
libspice.eknseg_c.argtypes = [c_int]
libspice.eknseg_c.restype = c_int
libspice.ekntab_c.argtypes = [POINTER(c_int)]
libspice.ekopn_c.argtypes = [c_char_p, c_char_p, c_int, POINTER(c_int)]
libspice.ekopr_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.ekops_c.argtypes = [POINTER(c_int)]
libspice.ekopw_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.ekpsel_c.argtypes = [c_char_p, c_int, c_int, c_int, POINTER(c_int), POINTER(c_int), POINTER(c_int),
                              POINTER(c_int), POINTER(c_int), c_void_p, c_void_p,
                              POINTER(c_bool), c_char_p]
#                              POINTER(stypes.SpiceEKDataType), POINTER(stypes.SpiceEKExprClass), c_void_p, c_void_p,
libspice.ekrcec_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, POINTER(c_int), c_void_p, POINTER(c_bool)]
libspice.ekrced_c.argtypes = [c_int, c_int, c_int, c_char_p, POINTER(c_int), POINTER(c_double), POINTER(c_bool)]
libspice.ekrcei_c.argtypes = [c_int, c_int, c_int, c_char_p, POINTER(c_int), POINTER(c_int), POINTER(c_bool)]
libspice.ekssum_c.argtypes = [c_int, c_int, POINTER(stypes.SpiceEKSegSum)]
libspice.ektnam_c.argtypes = [c_int, c_int, c_char_p]
libspice.ekucec_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, c_int, c_void_p, c_int]
libspice.ekuced_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, POINTER(c_double), c_int]
libspice.ekucei_c.argtypes = [c_int, c_int, c_int, c_char_p, c_int, POINTER(c_int), c_int]
libspice.ekuef_c.argtypes = [c_int]
libspice.el2cgv_c.argtypes = [POINTER(stypes.Ellipse), (c_double * 3), (c_double * 3), (c_double * 3)]
libspice.elemc_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.elemc_c.restype = c_bool
libspice.elemd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.elemd_c.restype = c_bool
libspice.elemi_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.elemi_c.restype = c_bool
libspice.eqncpv_c.argtypes = [c_double, c_double, (c_double * 9), c_double, c_double,
                              (c_double * 6)]
libspice.eqstr_c.argtypes = [c_char_p, c_char_p]
libspice.eqstr_c.restype = c_bool
libspice.erract_c.argtypes = [c_char_p, c_int, c_char_p]
libspice.errch_c.argtypes = [c_char_p, c_char_p]
libspice.errdev_c.argtypes = [c_char_p, c_int, c_char_p]
libspice.errdp_c.argtypes = [c_char_p, c_double]
libspice.errint_c.argtypes = [c_char_p, c_int]
libspice.errprt_c.argtypes = [c_char_p, c_int, c_char_p]
libspice.esrchc_c.argtypes = [c_char_p, c_int, c_int, c_void_p]
libspice.esrchc_c.restype = c_int
libspice.et2lst_c.argtypes = [c_double, c_int, c_double, c_char_p, c_int, c_int, POINTER(c_int), POINTER(c_int),
                              POINTER(c_int), c_char_p, c_char_p]
libspice.et2utc_c.argtypes = [c_double, c_char_p, c_int, c_int, c_char_p]
libspice.etcal_c.argtypes = [c_double, c_int, c_char_p]
libspice.eul2m_c.argtypes = [c_double, c_double, c_double, c_int, c_int, c_int, (c_double * 3) * 3]
libspice.eul2xf_c.argtypes = [(c_double * 6), c_int, c_int, c_int, (c_double * 6) * 6]
libspice.exists_c.argtypes = [c_char_p]
libspice.exists_c.restype = c_bool
libspice.expool_c.argtypes = [c_char_p, POINTER(c_bool)]

########################################################################################################################
# F

libspice.failed_c.argtypes = None
libspice.failed_c.restype = c_bool
libspice.fn2lun_.argtypes = [c_char_p, POINTER(c_int), c_int]
libspice.fovray_c.argtypes = [c_char_p, (c_double * 3), c_char_p, c_char_p, c_char_p,
                              POINTER(c_double), POINTER(c_bool)]
libspice.fovtrg_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p,
                              c_char_p, POINTER(c_double), POINTER(c_bool)]
libspice.frame_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.frinfo_c.argtypes = [c_int, POINTER(c_int), POINTER(c_int), POINTER(c_int), POINTER(c_bool)]
libspice.frmnam_c.argtypes = [c_int, c_int, c_char_p]
libspice.ftncls_c.argtypes = [c_int]
libspice.furnsh_c.argtypes = [c_char_p]

########################################################################################################################
# G

libspice.gcpool_c.argtypes = [c_char_p, c_int, c_int, c_int, POINTER(c_int), c_void_p, POINTER(c_bool)]
libspice.gdpool_c.argtypes = [c_char_p, c_int, c_int, POINTER(c_int), POINTER(c_double), POINTER(c_bool)]
libspice.georec_c.argtypes = [c_double, c_double, c_double, c_double, c_double, (c_double * 3)]
libspice.getcml_c.argtypes = [c_int, c_char_p]
libspice.getelm_c.argtypes = [c_int, c_int, c_void_p, POINTER(c_double), POINTER(c_double)]
libspice.getfat_c.argtypes = [c_char_p, c_int, c_int, c_char_p, c_char_p]
libspice.getfov_c.argtypes = [c_int, c_int, c_int, c_int, c_char_p, c_char_p, (c_double * 3), POINTER(c_int),
                              POINTER(c_double * 3)]
libspice.getmsg_c.argtypes = [c_char_p, c_int, c_char_p]
libspice.gfbail_c.restype  = c_bool
libspice.gfclrh_c.argtypes = None
libspice.gfdist_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_double, c_double, c_double, c_int,
                              POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
# libspice.gfevnt_c.argtypes = [c_double, c_double, c_double, c_double, c_bool, c_bool, c_double, c_char_p, c_int, c_int, c_char_p, c_double, c_double, c_double, c_bool, None, c_char_p, c_char_p, c_double, c_double, c_double, c_int, c_bool, c_bool, None, None]
# libspice.gffove_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_double, c_double, c_double, c_double,  c_double, c_bool, c_bool, c_double, c_bool, None,  c_char_p, c_char_p, c_double, c_double, c_double,  c_bool, c_bool, None, None]
libspice.gfinth_c.argtypes = [c_int]
libspice.gfilum_c.argtupes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, (c_double * 3), c_char_p, c_double, c_double, c_double, c_int, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
# libspice.gfocce_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_double, c_double, c_double, c_double, c_double, c_bool, c_bool, c_double, c_bool, None, c_char_p, c_char_p, c_double, c_double, c_double, c_bool, c_bool, None, None]
libspice.gfoclt_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p,
                              c_double, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfpa_c.argtypes   = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_double, c_double,
                              c_double, c_int, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfposc_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_double, c_double,
                              c_double, c_int, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfrefn_c.argtypes = [c_double, c_double, c_bool, c_bool, POINTER(c_double)]
libspice.gfrepf_c.argtypes = None
libspice.gfrepi_c.argtypes = [POINTER(stypes.SpiceCell), c_char_p, c_char_p]
libspice.gfrepu_c.argtypes = [c_double, c_double, c_double]
libspice.gfrfov_c.argtypes = [c_char_p, (c_double * 3), c_char_p, c_char_p, c_char_p, c_double,
                              POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfrr_c.argtypes   = [c_char_p, c_char_p, c_char_p, c_char_p, c_double, c_double, c_double, c_int,
                              POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfsep_c.argtypes  = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p,
                              c_double, c_double, c_double, c_int, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfsntc_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, (c_double * 3), c_char_p,
                              c_char_p, c_char_p, c_double, c_double, c_double, c_int, POINTER(stypes.SpiceCell),
                              POINTER(stypes.SpiceCell)]
libspice.gfsstp_c.argtypes = [c_double]
libspice.gfstep_c.argtypes = [c_double, POINTER(c_double)]
libspice.gfstol_c.argtypes = [c_double]
libspice.gfsubc_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_double,
                              c_double, c_double, c_int, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gftfov_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_double,
                              POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfudb_c.argtypes  = [callbacks.UDFUNS, callbacks.UDFUNB, c_double, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gfuds_c.argtypes  = [callbacks.UDFUNS, callbacks.UDFUNB, c_char_p, c_double, c_double, c_double, c_int, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.gipool_c.argtypes = [c_char_p, c_int, c_int, POINTER(c_int), POINTER(c_int), POINTER(c_bool)]
libspice.gnpool_c.argtypes = [c_char_p, c_int, c_int, c_int, POINTER(c_int), c_void_p, POINTER(c_bool)]

########################################################################################################################
# H

libspice.halfpi_c.restype = c_double
libspice.hrmint_c.argtypes = [c_int, POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.hx2dp_c.argtypes = [c_char_p, c_int, POINTER(c_double), POINTER(c_bool), c_char_p]

########################################################################################################################
# I

libspice.ident_c.argtypes  = [(c_double * 3) * 3]
libspice.illum_c.argtypes  = [c_char_p, c_double, c_char_p, c_char_p, (c_double * 3), POINTER(c_double),
                             POINTER(c_double), POINTER(c_double)]
libspice.ilumin_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3),
                              POINTER(c_double), (c_double * 3), POINTER(c_double), POINTER(c_double),
                              POINTER(c_double)]
libspice.illumf_c.argtypes = [c_char_p, c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3), POINTER(c_double), (c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_bool), POINTER(c_bool)]
libspice.illumg_c.argtypes = [c_char_p, c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3), POINTER(c_double), (c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.inedpl_c.argtypes = [c_double, c_double, c_double, POINTER(stypes.Plane), POINTER(stypes.Ellipse),
                              POINTER(c_bool)]
libspice.inelpl_c.argtypes = [POINTER(stypes.Ellipse), POINTER(stypes.Plane), POINTER(c_int), (c_double * 3),
                              (c_double * 3)]

libspice.insrtc_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.insrtd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.insrti_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.inter_c.argtypes  = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.inrypl_c.argtypes = [(c_double * 3), (c_double * 3), POINTER(stypes.Plane), POINTER(c_int), (c_double * 3)]
libspice.intmax_c.restype  = c_int
libspice.intmin_c.restype  = c_int
libspice.invert_c.argtypes = [(c_double * 3) * 3, (c_double * 3) * 3]
libspice.invort_c.argtypes = [(c_double * 3) * 3, (c_double * 3) * 3]
libspice.isordv_c.argtypes = [POINTER(c_int), c_int]
libspice.isordv_c.restype  = c_bool
libspice.isrchc_c.argtypes = [c_char_p, c_int, c_int, c_void_p]
libspice.isrchc_c.restype  = c_int
libspice.isrchd_c.argtypes = [c_double, c_int, POINTER(c_double)]
libspice.isrchd_c.restype  = c_int
libspice.isrchi_c.argtypes = [c_int, c_int, POINTER(c_int)]
libspice.isrchi_c.restype  = c_int
libspice.isrot_c.argtypes  = [(c_double * 3) * 3, c_double, c_double]
libspice.isrot_c.restype   = c_bool
libspice.iswhsp_c.argtypes = [c_char_p]
libspice.iswhsp_c.restype  = c_bool
########################################################################################################################
# J

libspice.j1900_c.restype = c_double
libspice.j1950_c.restype = c_double
libspice.j2000_c.restype = c_double
libspice.j2100_c.restype = c_double
libspice.jyear_c.restype = c_double

########################################################################################################################
# K
libspice.kclear_c.restype  = None
libspice.kdata_c.argtypes  = [c_int, c_char_p, c_int, c_int, c_int, c_char_p, c_char_p, c_char_p, POINTER(c_int),
                             POINTER(c_bool)]
libspice.kinfo_c.argtypes  = [c_char_p, c_int, c_int, c_char_p, c_char_p, POINTER(c_int), POINTER(c_bool)]
libspice.ktotal_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.kplfrm_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.kxtrct_c.argtypes = [c_char_p, c_int, c_void_p, c_int, c_int, c_int, c_char_p, POINTER(c_bool), c_char_p]

########################################################################################################################
# L
libspice.lastnb_c.argtypes = [c_char_p]
libspice.lastnb_c.restype  = c_int
libspice.latcyl_c.argtypes = [c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.latrec_c.argtypes = [c_double, c_double, c_double, (c_double) * 3]
libspice.latsph_c.argtypes = [c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.latsrf_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_int, c_void_p, c_void_p]
libspice.lcase_c.argtypes  = [c_char_p, c_int, c_char_p]
libspice.ldpool_c.argtypes = [c_char_p]
libspice.lgrind_c.argtypes = [c_int, POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double)]
libspice.limbpt_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, c_char_p, (c_double * 3), c_double, c_int, c_double, c_double, c_int, POINTER(c_int), c_void_p, POINTER(c_double), c_void_p]
libspice.lmpool_c.argtypes = [c_void_p, c_int, c_int]
libspice.lparse_c.argtypes = [c_char_p, c_char_p, c_int, c_int, POINTER(c_int), c_void_p]
libspice.lparsm_c.argtypes = [c_char_p, c_char_p, c_int, c_int, POINTER(c_int), c_void_p]
libspice.lparss_c.argtypes = [c_char_p, c_char_p, POINTER(stypes.SpiceCell)]
libspice.lspcn_c.argtypes  = [c_char_p, c_double, c_char_p]
libspice.lspcn_c.restype   = c_double
libspice.ltime_c.argtypes  = [c_double, c_int, c_char_p, c_int, POINTER(c_double), POINTER(c_double)]
libspice.lstlec_c.argtypes = [c_char_p, c_int, c_int, c_void_p]
libspice.lstlec_c.restype  = c_int
libspice.lstled_c.argtypes = [c_double, c_int, POINTER(c_double)]
libspice.lstled_c.restype  = c_int
libspice.lstlei_c.argtypes = [c_int, c_int, POINTER(c_int)]
libspice.lstlei_c.restype  = c_int
libspice.lstltc_c.argtypes = [c_char_p, c_int, c_int, c_void_p]
libspice.lstltc_c.restype  = c_int
libspice.lstltd_c.argtypes = [c_double, c_int, POINTER(c_double)]
libspice.lstltd_c.restype  = c_int
libspice.lstlti_c.argtypes = [c_int, c_int, POINTER(c_int)]
libspice.lstlti_c.restype  = c_int
libspice.lx4dec_c.argtypes = [c_char_p, c_int, POINTER(c_int), POINTER(c_int)]
libspice.lx4num_c.argtypes = [c_char_p, c_int, POINTER(c_int), POINTER(c_int)]
libspice.lx4sgn_c.argtypes = [c_char_p, c_int, POINTER(c_int), POINTER(c_int)]
libspice.lx4uns_c.argtypes = [c_char_p, c_int, POINTER(c_int), POINTER(c_int)]
libspice.lxqstr_c.argtypes = [c_char_p, c_char, c_int, POINTER(c_int), POINTER(c_int)]
########################################################################################################################
# M

libspice.m2eul_c.argtypes = [(c_double * 3) * 3, c_int, c_int, c_int, POINTER(c_double), POINTER(c_double),
                             POINTER(c_double)]
libspice.m2q_c.argtypes = [(c_double * 3) * 3, (c_double * 4)]
libspice.matchi_c.argtypes = [c_char_p, c_char_p, c_char, c_char]
libspice.matchi_c.restype = c_bool
libspice.matchw_c.argtypes = [c_char_p, c_char_p, c_char, c_char]
libspice.matchw_c.restype = c_bool
libspice.maxd_c.restype = c_double
libspice.mequ_c.argtypes = [(c_double * 3) * 3, (c_double * 3) * 3, ]
libspice.mequg_c.argtypes = [c_void_p, c_int, c_int, c_void_p]
libspice.mtxm_c.argtypes = [(c_double * 3) * 3, (c_double * 3) * 3, (c_double * 3) * 3]
libspice.mtxmg_c.argtypes = [c_void_p, c_void_p, c_int, c_int, c_int, c_void_p]
libspice.mtxv_c.argtypes = [(c_double * 3) * 3, (c_double * 3), (c_double * 3)]
libspice.mtxvg_c.argtypes = [c_void_p, c_void_p, c_int, c_int, c_void_p]
libspice.mxm_c.argtypes = [(c_double * 3) * 3, (c_double * 3) * 3, (c_double * 3) * 3]
libspice.mxmg_c.argtypes = [c_void_p, c_void_p, c_int, c_int, c_int, c_void_p]
libspice.mxmt_c.argtypes = [(c_double * 3) * 3, (c_double * 3) * 3, (c_double * 3) * 3]
libspice.mxmtg_c.argtypes = [c_void_p, c_void_p, c_int, c_int, c_int, c_void_p]
libspice.mxv_c.argtypes = [(c_double * 3) * 3, (c_double * 3), (c_double * 3)]
libspice.mxvg_c.argtypes = [c_void_p, c_void_p, c_int, c_int, c_void_p]

########################################################################################################################
# N

libspice.namfrm_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.ncpos_c.argtypes = [c_char_p, c_char_p, c_int]
libspice.ncpos_c.restype = c_int
libspice.ncposr_c.argtypes = [c_char_p, c_char_p, c_int]
libspice.ncposr_c.restype = c_int
libspice.nearpt_c.argtypes = [(c_double * 3), c_double, c_double, c_double, (c_double * 3), POINTER(c_double)]
libspice.npedln_c.argtypes = [c_double, c_double, c_double, (c_double * 3), (c_double * 3), (c_double * 3),
                              POINTER(c_double)]
libspice.npelpt_c.argtypes = [(c_double * 3), POINTER(stypes.Ellipse), (c_double * 3), POINTER(c_double)]
libspice.nplnpt_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3), (c_double * 3), POINTER(c_double)]
libspice.nvc2pl_c.argtypes = [(c_double * 3), c_double, POINTER(stypes.Plane)]
libspice.nvp2pl_c.argtypes = [(c_double * 3), (c_double * 3), POINTER(stypes.Plane)]

########################################################################################################################
# O
libspice.occult_c.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p,
                              c_char_p, c_char_p, c_char_p, c_double,
                              POINTER(c_int)]
libspice.ordc_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.ordc_c.restype = c_int
libspice.ordd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.ordd_c.restype = c_int
libspice.ordi_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.ordi_c.restype = c_int
libspice.orderc_c.argtypes = [c_int, c_void_p, c_int, POINTER(c_int)]
libspice.orderd_c.argtypes = [POINTER(c_double), c_int, POINTER(c_int)]
libspice.orderi_c.argtypes = [POINTER(c_int), c_int, POINTER(c_int)]
libspice.oscelt_c.argtypes = [c_double * 6, c_double, c_double, c_double * 8]
libspice.oscltx_c.argtypes = [(c_double * 6), c_double, c_double, POINTER(c_double)]

########################################################################################################################
# P

libspice.pckcls_c.argtypes = [c_int]
libspice.pckcov_c.argtypes = [c_char_p, c_int, POINTER(stypes.SpiceCell)]
libspice.pckfrm_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.pcklof_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.pckopn_c.argtypes = [c_char_p, c_char_p, c_int, POINTER(c_int)]
libspice.pckuof_c.argtypes = [c_int]
libspice.pckw02_c.argtypes = [c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, c_int, c_int, POINTER(c_double), c_double]
libspice.pcpool_c.argtypes = [c_char_p, c_int, c_int, c_void_p]
libspice.pdpool_c.argtypes = [c_char_p, c_int, POINTER(c_double)]
libspice.pipool_c.argtypes = [c_char_p, c_int, POINTER(c_int)]
libspice.pgrrec_c.argtypes = [c_char_p, c_double, c_double, c_double, c_double, c_double, (c_double * 3)]
libspice.phaseq_c.argtypes = [c_double, c_char_p, c_char_p, c_char_p, c_char_p]
libspice.phaseq_c.restype = c_double
libspice.pi_c.restype = c_double
libspice.pjelpl_c.argtypes = [POINTER(stypes.Ellipse), POINTER(stypes.Plane), POINTER(stypes.Ellipse)]
libspice.pl2nvc_c.argtypes = [POINTER(stypes.Plane), (c_double * 3), POINTER(c_double)]
libspice.pl2nvp_c.argtypes = [POINTER(stypes.Plane), (c_double * 3), (c_double * 3)]
libspice.pl2psv_c.argtypes = [POINTER(stypes.Plane), (c_double * 3), (c_double * 3), (c_double * 3)]
libspice.pltar_c.argtypes = [c_int, c_void_p, c_int, c_void_p]
libspice.pltar_c.restype = c_double
libspice.pltexp_c.argtypes = [(c_double * 3) * 3, c_double, (c_double * 3) * 3]
libspice.pltnp_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3), (c_double * 3), (c_double * 3), POINTER(c_double)]
libspice.pltnrm_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3), (c_double * 3)]
libspice.pltvol_c.argtypes = [c_int, c_void_p, c_int, c_void_p]
libspice.pltvol_c.restype = c_double
libspice.polyds_c.argtype = [POINTER(c_double), c_int, c_int, c_double, POINTER(c_double)]
libspice.pos_c.argtypes = [c_char_p, c_char_p, c_int]
libspice.pos_c.restype = c_int
libspice.posr_c.argtypes = [c_char_p, c_char_p, c_int]
libspice.posr_c.restype = c_int
# libspice.prefix_c.argtypes = [c_char_p, c_int, c_int, c_char_p]
libspice.prop2b_c.argtypes = [c_double, (c_double * 6), c_double, (c_double * 6)]
libspice.prsdp_c.argtypes = [c_char_p, POINTER(c_double)]
libspice.prsint_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.psv2pl_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3), POINTER(stypes.Plane)]
libspice.putcml_c.argtypes = [c_int, c_char_p]
libspice.pxform_c.argtypes = [c_char_p, c_char_p, c_double, (c_double * 3) * 3]
libspice.pxfrm2_c.argtypes = [c_char_p, c_char_p, c_double, c_double, (c_double * 3) * 3]

########################################################################################################################
# Q

libspice.q2m_c.argtypes = [c_double * 4, (c_double * 3) * 3]
libspice.qcktrc_c.argtypes = [c_int, c_char_p]
libspice.qdq2av_c.argtypes = [c_double * 4, c_double * 4, c_double * 3]
libspice.qxq_c.argtypes = [c_double * 4, c_double * 4, c_double * 4]

########################################################################################################################
# R
libspice.radrec_c.argtypes = [c_double, c_double, c_double, (c_double * 3)]
libspice.rav2xf_c.argtypes = [(c_double * 3) * 3, (c_double * 3), (c_double * 6) * 6]
libspice.raxisa_c.argtypes = [(c_double * 3) * 3, (c_double * 3), POINTER(c_double)]
libspice.rdtext_c.argtypes = [c_char_p, c_int, c_char_p, POINTER(c_bool)]
libspice.reccyl_c.argtypes = [(c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.recgeo_c.argtypes = [(c_double * 3), c_double, c_double, POINTER(c_double), POINTER(c_double),
                              POINTER(c_double)]
libspice.reclat_c.argtypes = [(c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.recpgr_c.argtypes = [c_char_p, (c_double * 3), c_double, c_double, POINTER(c_double), POINTER(c_double),
                              POINTER(c_double)]
libspice.recrad_c.argtypes = [(c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.recsph_c.argtypes = [(c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.reordc_c.argtypes = [POINTER(c_int), c_int, c_int, c_void_p]
libspice.reordd_c.argtypes = [POINTER(c_int), c_int, POINTER(c_double)]
libspice.reordi_c.argtypes = [POINTER(c_int), c_int, POINTER(c_int)]
libspice.reordl_c.argtypes = [POINTER(c_int), c_int, POINTER(c_bool)]
libspice.removc_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.removd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.removi_c.argtypes = [c_int, POINTER(stypes.SpiceCell)]
libspice.repmc_c.argtypes = [c_char_p, c_char_p, c_char_p, c_int, c_char_p]
libspice.repmct_c.argtypes = [c_char_p, c_char_p, c_int, c_char, c_int, c_char_p]
libspice.repmd_c.argtypes = [c_char_p, c_char_p, c_double, c_int, c_int, c_char_p]
libspice.repmf_c.argtypes = [c_char_p, c_char_p, c_double, c_int, c_char, c_int, c_char_p]
libspice.repmi_c.argtypes = [c_char_p, c_char_p, c_int, c_int, c_char_p]
libspice.repmot_c.argtypes = [c_char_p, c_char_p, c_int, c_char, c_int, c_char_p]
libspice.reset_c.argtypes = None
libspice.return_c.argtypes = None
libspice.return_c.restype = c_bool
libspice.rotate_c.argtypes = [c_double, c_int, (c_double * 3) * 3]
libspice.rotmat_c.argtypes = [(c_double * 3) * 3, c_double, c_int, (c_double * 3) * 3]
libspice.rotvec_c.argtypes = [(c_double * 3), c_double, c_int, (c_double * 3)]
libspice.rpd_c.restype = c_double
libspice.rquad_c.argtypes = [c_double, c_double, c_double, (c_double * 2), (c_double * 2)]

########################################################################################################################
# S

libspice.saelgv_c.argtypes = [(c_double * 3), (c_double * 3), (c_double * 3), (c_double * 3)]
libspice.scard_c.argtypes  = [c_int, POINTER(stypes.SpiceCell)]
libspice.scdecd_c.argtypes = [c_int, c_double, c_int, c_char_p]
libspice.sce2c_c.argtypes  = [c_int, c_double, POINTER(c_double)]
libspice.sce2s_c.argtypes  = [c_int, c_double, c_int, c_char_p]
libspice.sce2t_c.argtypes  = [c_int, c_double, POINTER(c_double)]
libspice.scencd_c.argtypes = [c_int, c_char_p, POINTER(c_double)]
libspice.scfmt_c.argtypes  = [c_int, c_double, c_int, c_char_p]
libspice.scpart_c.argtypes = [c_int, POINTER(c_int), POINTER(c_double), POINTER(c_double)]
libspice.scs2e_c.argtypes  = [c_int, c_char_p, POINTER(c_double)]
libspice.sct2e_c.argtypes  = [c_int, c_double, POINTER(c_double)]
libspice.sctiks_c.argtypes = [c_int, c_char_p, POINTER(c_double)]
libspice.sdiff_c.argtypes  = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.set_c.argtypes    = [POINTER(stypes.SpiceCell), c_char_p, POINTER(stypes.SpiceCell)]
libspice.set_c.restype     = c_bool
libspice.setmsg_c.argtypes = [c_char_p]
libspice.shellc_c.argtypes = [c_int, c_int, c_void_p]
libspice.shelld_c.argtypes = [c_int, POINTER(c_double)]
libspice.shelli_c.argtypes = [c_int, POINTER(c_int)]
libspice.sigerr_c.argtypes = [c_char_p]
libspice.sincpt_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, c_char_p, (c_double * 3),
                              (c_double * 3), POINTER(c_double), (c_double * 3), POINTER(c_bool)]
libspice.spd_c.restype     = c_double
libspice.sphcyl_c.argtypes = [c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.sphlat_c.argtypes = [c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.sphrec_c.argtypes = [c_double, c_double, c_double, (c_double * 3)]
libspice.spk14a_c.argtypes = [c_int, c_int, POINTER(c_double), POINTER(c_double)]
libspice.spk14b_c.argtypes = [c_int, c_char_p, c_int, c_int, c_char_p, c_double, c_double, c_int]
libspice.spk14e_c.argtypes = [c_int]
libspice.spkacs_c.argtypes = [c_int, c_double, c_char_p, c_char_p, c_int, (c_double * 6), POINTER(c_double),
                              POINTER(c_double)]
libspice.spkapo_c.argtypes = [c_int, c_double, c_char_p, (c_double * 6), c_char_p, (c_double * 3), POINTER(c_double)]
libspice.spkapp_c.argtypes = [c_int, c_double, c_char_p, (c_double * 6), c_char_p, (c_double * 6), POINTER(c_double)]
libspice.spkaps_c.argtypes = [c_int, c_double, c_char_p, c_char_p, (c_double * 6), (c_double * 6), (c_double * 6),
                              POINTER(c_double), POINTER(c_double)]
libspice.spkcls_c.argtypes = [c_int]
libspice.spkcov_c.argtypes = [c_char_p, c_int, POINTER(stypes.SpiceCell)]
libspice.spkcpo_c.argtypes = [c_char_p, c_double, c_char_p, c_char_p, c_char_p,
                              (c_double * 3), c_char_p, c_char_p, (c_double * 6),
                              POINTER(c_double)]
libspice.spkcpt_c.argtypes = [(c_double * 3), c_char_p, c_char_p, c_double, c_char_p,
                              c_char_p, c_char_p, c_char_p, (c_double * 6),
                              POINTER(c_double)]
libspice.spkcvo_c.argtypes = [c_char_p, c_double, c_char_p, c_char_p, c_char_p,
                              (c_double * 6), c_double, c_char_p, c_char_p,
                              (c_double * 6), POINTER(c_double)]
libspice.spkcvt_c.argtypes = [(c_double * 6), c_double, c_char_p, c_char_p, c_double,
                              c_char_p, c_char_p, c_char_p, c_char_p, (c_double * 6),
                              POINTER(c_double)]
libspice.spkez_c.argtypes  = [c_int, c_double, c_char_p, c_char_p, c_int, (c_double * 6), POINTER(c_double)]
libspice.spkezp_c.argtypes = [c_int, c_double, c_char_p, c_char_p, c_int, (c_double * 3), POINTER(c_double)]
libspice.spkezr_c.argtypes = [c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 6), POINTER(c_double)]
libspice.spkgeo_c.argtypes = [c_int, c_double, c_char_p, c_int, (c_double * 6), POINTER(c_double)]
libspice.spkgps_c.argtypes = [c_int, c_double, c_char_p, c_int, (c_double * 3), POINTER(c_double)]
libspice.spklef_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.spkltc_c.argtypes = [c_int, c_double, c_char_p, c_char_p, (c_double * 6), (c_double * 6), POINTER(c_double),
                              POINTER(c_double)]
libspice.spkobj_c.argtypes = [c_char_p, POINTER(stypes.SpiceCell)]
libspice.spkopa_c.argtypes = [c_char_p, POINTER(c_int)]
libspice.spkopn_c.argtypes = [c_char_p, c_char_p, c_int, POINTER(c_int)]
libspice.spkpds_c.argtypes = [c_int, c_int, c_char_p, c_int, c_double, c_double, (c_double * 5)]
libspice.spkpos_c.argtypes = [c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3), POINTER(c_double)]
libspice.spkpvn_c.argtypes = [c_int, (c_double * 5), c_double, POINTER(c_int),
                              (c_double * 6), POINTER(c_int)]
libspice.spksfs_c.argtypes = [c_int, c_double, c_int, POINTER(c_int),
                              (c_double * 5), c_char_p, POINTER(c_bool)]
libspice.spkssb_c.argtypes = [c_int, c_double, c_char_p, (c_double * 6)]
libspice.spksub_c.argtypes = [c_int, (c_double * 5), c_char_p, c_double, c_double, c_int]
libspice.spkuds_c.argtypes = [(c_double * 5), POINTER(c_int), POINTER(c_int), POINTER(c_int), POINTER(c_int),
                              POINTER(c_double), POINTER(c_double), POINTER(c_int), POINTER(c_int)]
libspice.spkuef_c.argtypes = [c_int]
libspice.spkw02_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, c_int, c_int,
                              POINTER(c_double), c_double]
libspice.spkw03_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, c_int, c_int,
                              POINTER(c_double), c_double]
libspice.spkw05_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, c_int,
                              POINTER(c_double * 6), POINTER(c_double)]
libspice.spkw08_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_int, c_int,
                              POINTER(c_double * 6), c_double, c_double]
libspice.spkw09_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_int, c_int,
                              POINTER(c_double * 6), POINTER(c_double)]
libspice.spkw10_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, (c_double * 8), c_int,
                              POINTER(c_double), POINTER(c_double)]
libspice.spkw12_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_int, c_int,
                              POINTER(c_double * 6), c_double, c_double]
libspice.spkw13_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_int, c_int,
                              POINTER(c_double * 6), POINTER(c_double)]
libspice.spkw15_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, (c_double * 3),
                              (c_double * 3), c_double, c_double, c_double, (c_double * 3), c_double, c_double,
                              c_double]
libspice.spkw17_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, (c_double * 9),
                              c_double, c_double]
libspice.spkw18_c.argtypes = [c_int, c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_int, c_int, c_void_p, POINTER(c_double)]
libspice.spkw20_c.argtypes = [c_int, c_int, c_int, c_char_p, c_double, c_double, c_char_p, c_double, c_int, c_int, POINTER(c_double), c_double, c_double, c_double, c_double]
libspice.srfc2s_c.argtypes = [c_int, c_int, c_int, c_char_p, POINTER(c_bool)]
libspice.srfcss_c.argtypes = [c_int, c_char_p, c_int, c_char_p, POINTER(c_bool)]
libspice.srfnrm_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_int, c_void_p, c_void_p]
libspice.srfrec_c.argtypes = [c_int, c_double, c_double, (c_double * 3)]
libspice.srfs2c_c.argtypes = [c_char_p, c_char_p, POINTER(c_int), POINTER(c_bool)]
libspice.srfscc_c.argtypes = [c_char_p, c_int, POINTER(c_int), POINTER(c_bool)]
libspice.size_c.argtypes   = [POINTER(stypes.SpiceCell)]
libspice.size_c.restype    = c_int
libspice.srfxpt_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3),
                              (c_double * 3), POINTER(c_double), POINTER(c_double), (c_double * 3), POINTER(c_bool)]
libspice.ssize_c.argtypes  = [c_int, POINTER(stypes.SpiceCell)]
libspice.stelab_c.argtypes = [(c_double * 3), (c_double * 3)]
libspice.stpool_c.argtypes = [c_char_p, c_int, c_char_p, c_int, c_char_p, POINTER(c_int), POINTER(c_bool)]
libspice.str2et_c.argtypes = [c_char_p, POINTER(c_double)]
libspice.subpnt_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3),
                              POINTER(c_double), (c_double * 3)]
libspice.subpt_c.argtypes  = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, (c_double * 3), POINTER(c_double)]
libspice.subslr_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, (c_double * 3),
                              POINTER(c_double), (c_double * 3)]
libspice.subsol_c.argtypes = [c_char_p, c_char_p, c_double, c_char_p, c_char_p, (c_double * 3)]
libspice.sumai_c.argtypes  = [POINTER(c_int), c_int]
libspice.sumai_c.restype   = c_int
libspice.sumad_c.argtypes  = [POINTER(c_double), c_int]
libspice.sumad_c.restype   = c_double
libspice.surfnm_c.argtypes = [c_double, c_double, c_double, (c_double * 3), (c_double * 3)]
libspice.surfpt_c.argtypes = [(c_double * 3), (c_double * 3), c_double, c_double,
                              c_double, (c_double * 3), POINTER(c_bool)]
libspice.surfpv_c.argtypes = [(c_double * 6), (c_double * 6), c_double, c_double, c_double, (c_double * 6),
                              POINTER(c_bool)]
libspice.swpool_c.argtypes = [c_char_p, c_int, c_int, c_void_p]
libspice.sxform_c.argtypes = [c_char_p, c_char_p, c_double, (c_double * 6) * 6]
libspice.szpool_c.argtypes = [c_char_p, POINTER(c_int), POINTER(c_bool)]

########################################################################################################################
# T

libspice.termpt_c.argtypes = [c_char_p, c_char_p, c_char_p, c_double, c_char_p, c_char_p, c_char_p, c_char_p, (c_double * 3), c_double, c_int, c_double, c_double, c_int, POINTER(c_int), c_void_p, POINTER(c_double), c_void_p]
libspice.timdef_c.argtypes = [c_char_p, c_char_p, c_int, c_char_p]
libspice.timout_c.argtypes = [c_double, c_char_p, c_int, c_char_p]
libspice.tipbod_c.argtypes = [c_char_p, c_int, c_double, (c_double * 3) * 3]
libspice.tisbod_c.argtypes = [c_char_p, c_int, c_double, (c_double * 6) * 6]
libspice.tkvrsn_c.argtypes = [c_char_p]
libspice.tkvrsn_c.restype  = c_char_p
libspice.tparse_c.argtypes = [c_char_p, c_int, POINTER(c_double), c_char_p]
libspice.tpictr_c.argtypes = [c_char_p, c_int, c_int, c_char_p, POINTER(c_bool), c_char_p]
libspice.trace_c.argtypes  = [(c_double * 3) * 3]
libspice.trace_c.restype   = c_double
libspice.trcdep_c.argtypes = [POINTER(c_int)]
libspice.trcnam_c.argtypes = [c_int, c_int, c_char_p]
libspice.trcoff_c.argtypes = None
libspice.tsetyr_c.argtypes = [c_int]
libspice.twopi_c.restype   = c_double
libspice.twovec_c.argtypes = [(c_double * 3), c_int, (c_double * 3), c_int, (c_double * 3) * 3]
libspice.txtopn_.argtypes   = [c_char_p, POINTER(c_int), c_int]
libspice.tyear_c.restype   = c_double

########################################################################################################################
# U

libspice.ucase_c.argtypes  = [c_char_p, c_int, c_char_p]
libspice.ucrss_c.argtypes  = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.uddc_c.argtypes   = [callbacks.UDFUNS, c_double, c_double, POINTER(c_bool)]
libspice.uddc_c.restype    = None
libspice.uddf_c.argtypes   = [callbacks.UDFUNS, c_double, c_double, POINTER(c_double)]
libspice.uddf_c.restype    = None
libspice.udf_c.argtypes    = [c_double, POINTER(c_double)]
libspice.udf_c.restype     = None
libspice.union_c.argtypes  = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.unitim_c.argtypes = [c_double, c_char_p, c_char_p]
libspice.unitim_c.restype  = c_double
libspice.unload_c.argtypes = [c_char_p]
libspice.unorm_c.argtypes  = [(c_double * 3), (c_double * 3), POINTER(c_double)]
libspice.unormg_c.argtypes = [POINTER(c_double), c_int, POINTER(c_double), POINTER(c_double)]
libspice.utc2et_c.argtypes = [c_char_p, POINTER(c_double)]
########################################################################################################################
# V

libspice.vadd_c.argtypes   = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.vaddg_c.argtypes  = [POINTER(c_double), POINTER(c_double), c_int, POINTER(c_double)]
libspice.valid_c.argtypes  = [c_int, c_int, POINTER(stypes.SpiceCell)]
libspice.vcrss_c.argtypes  = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.vdist_c.argtypes  = [(c_double * 3), (c_double * 3)]
libspice.vdist_c.restype   = c_double
libspice.vdistg_c.argtypes = [POINTER(c_double), POINTER(c_double), c_int]
libspice.vdistg_c.restype  = c_double
libspice.vdot_c.argtypes   = [(c_double * 3), (c_double * 3)]
libspice.vdot_c.restype    = c_double
libspice.vdotg_c.argtypes  = [POINTER(c_double), POINTER(c_double), c_int]
libspice.vdotg_c.restype   = c_double
libspice.vequ_c.argtypes   = [(c_double * 3), (c_double * 3)]
libspice.vequg_c.argtypes  = [POINTER(c_double), c_int, POINTER(c_double)]
libspice.vhat_c.argtypes   = [(c_double * 3), (c_double * 3)]
libspice.vhatg_c.argtypes  = [POINTER(c_double), c_int, POINTER(c_double)]
libspice.vlcom_c.argtypes  = [c_double, (c_double * 3), c_double, (c_double * 3), (c_double * 3)]
libspice.vlcom3_c.argtypes = [c_double, (c_double * 3), c_double, (c_double * 3), c_double, (c_double * 3),
                              (c_double * 3)]
libspice.vlcomg_c.argtypes = [c_int, c_double, POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double)]
libspice.vminug_c.argtypes = [POINTER(c_double), c_int, POINTER(c_double)]
libspice.vminus_c.argtypes = [(c_double * 3), (c_double * 3)]
libspice.vnorm_c.restype   = c_double
libspice.vnorm_c.argstype  = [stypes.emptyDoubleVector(3)]
libspice.vnormg_c.restype  = c_double
libspice.vnormg_c.argstype = [POINTER(c_double), c_int]
libspice.vpack_c.argtypes  = [c_double, c_double, c_double, (c_double * 3)]
libspice.vperp_c.argtypes  = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.vprjp_c.argtypes  = [(c_double * 3), POINTER(stypes.Plane), (c_double * 3)]
libspice.vprjpi_c.argtypes = [(c_double * 3), POINTER(stypes.Plane), POINTER(stypes.Plane), (c_double * 3),
                              POINTER(c_bool)]
libspice.vproj_c.argtypes  = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.vrelg_c.argtypes  = [POINTER(c_double), POINTER(c_double), c_int]
libspice.vrelg_c.restype   = c_double
libspice.vrel_c.argtypes   = [(c_double * 3), (c_double * 3)]
libspice.vrel_c.restype    = c_double
libspice.vrotv_c.argtypes  = [(c_double * 3), (c_double * 3), c_double, (c_double * 3)]
libspice.vscl_c.argtypes   = [c_double, (c_double * 3), (c_double * 3)]
libspice.vsclg_c.argtypes  = [c_double, POINTER(c_double), c_int, POINTER(c_double)]
libspice.vsep_c.argtypes   = [(c_double * 3), (c_double * 3)]
libspice.vsep_c.restype    = c_double
libspice.vsepg_c.argtypes  = [POINTER(c_double), POINTER(c_double), c_int]
libspice.vsepg_c.restype   = c_double
libspice.vsub_c.argtypes   = [(c_double * 3), (c_double * 3), (c_double * 3)]
libspice.vsubg_c.argtypes  = [POINTER(c_double), POINTER(c_double), c_int, POINTER(c_double)]
libspice.vtmv_c.argtypes   = [(c_double * 3), (c_double * 3) * 3, (c_double * 3)]
libspice.vtmv_c.restype    = c_double
libspice.vtmvg_c.argtypes  = [POINTER(c_double), c_void_p, POINTER(c_double), c_int, c_int]
libspice.vtmvg_c.restype   = c_double
libspice.vupack_c.argtypes = [(c_double * 3), POINTER(c_double), POINTER(c_double), POINTER(c_double)]
libspice.vzero_c.argtypes  = [(c_double * 3)]
libspice.vzero_c.restype   = c_bool
libspice.vzerog_c.argtypes = [POINTER(c_double), c_int]
libspice.vzerog_c.restype  = c_bool

########################################################################################################################
# W
libspice.wncard_c.argtypes = [POINTER(stypes.SpiceCell)]
libspice.wncard_c.restype  = c_int
libspice.wncomd_c.argtypes = [c_double, c_double, POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.wncond_c.argtypes = [c_double, c_double, POINTER(stypes.SpiceCell)]
libspice.wndifd_c.argtypes = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.wnelmd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.wnelmd_c.restype  = c_bool
libspice.wnexpd_c.argtypes = [c_double, c_double, POINTER(stypes.SpiceCell)]
libspice.wnextd_c.argtypes = [c_char, POINTER(stypes.SpiceCell)]
libspice.wnfetd_c.argtypes = [POINTER(stypes.SpiceCell), c_int, POINTER(c_double), POINTER(c_double)]
libspice.wnfild_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.wnfltd_c.argtypes = [c_double, POINTER(stypes.SpiceCell)]
libspice.wnincd_c.argtypes = [c_double, c_double, POINTER(stypes.SpiceCell)]
libspice.wnincd_c.restype  = c_bool
libspice.wninsd_c.argtypes = [c_double, c_double, POINTER(stypes.SpiceCell)]
libspice.wnintd_c.argtypes = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.wnreld_c.argtypes = [POINTER(stypes.SpiceCell), c_char_p, POINTER(stypes.SpiceCell)]
libspice.wnreld_c.restype  = c_bool
libspice.wnsumd_c.argtypes = [POINTER(stypes.SpiceCell), POINTER(c_double), POINTER(c_double), POINTER(c_double),
                              POINTER(c_int), POINTER(c_int)]
libspice.wnunid_c.argtypes = [POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell), POINTER(stypes.SpiceCell)]
libspice.wnvald_c.argtypes = [c_int, c_int, POINTER(stypes.SpiceCell)]
libspice.writln_.argtypes = [c_char_p, POINTER(c_int), c_int]
########################################################################################################################
# X

libspice.xf2eul_c.argtypes = [(c_double * 6) * 6, c_int, c_int, c_int, (c_double * 6), POINTER(c_bool)]
libspice.xf2rav_c.argtypes = [(c_double * 6) * 6, (c_double * 3) * 3, (c_double * 3)]
libspice.xfmsta_c.argtypes = [(c_double * 6), c_char_p, c_char_p, c_char_p, (c_double * 6)]
libspice.xpose_c.argtypes  = [(c_double * 3) * 3, (c_double * 3) * 3]
libspice.xpose6_c.argtypes = [(c_double * 6) * 6, (c_double * 6) * 6]
libspice.xposeg_c.argtypes = [c_void_p, c_int, c_int, c_void_p]
########################################################################################################################
# Z

libspice.zzgetcml_c.argtypes = [c_int, c_char_p, c_bool]
libspice.zzgfsavh_c.argtypes = [c_bool]
#libspice.zzsynccl_c.argtypes = [None]
